/* ============================================================================
 * DCOPFJ : An open-source Java solver for DC optimal power flow (OPF) problems
 * ============================================================================
 * 
 * (C) Copyright 2007, by Junjie Sun and Leigh Tesfatsion
 *
 * Version 1.1
 *
 * Project site:  http://www.econ.iastate.edu/tesfatsi/DCOPFJHome.htm
 *
 * LICENSING TERMS
 * The DCOPFJ package is licensed by the copyright holders (Junjie Sun and 
 * Leigh Tesfatsion) as free open-source software under the terms of the GNU
 * General Public License (GPL). Anyone who is interested is allowed to view, 
 * modify, and/or improve upon the code used to produce this package, but any 
 * software generated using all or part of this code must be released as free 
 * open-source software in turn. The GNU GPL can be viewed in its entirety as 
 * in the following site: http://www.gnu.org/licenses/gpl.html
 */

package dcopfj;

// DCOPFJ implementation Version 1.1
// DCOPFJ.java
// DC Optimal Power Flow (OPF) solver
//
// DCOPFJ formulates the needed matrices and vectors for QuadProgJ
//
// DC-OPF paper: Junjie Sun and Leigh Tesfatsion, (2006) "DC OPF Formulation
//     and Solution Using QuadProgJ", ISU Econ Working Paper Series #06014

import cern.colt.matrix.*;
import cern.colt.matrix.impl.*;
import quadprogj.*;

/**
 * Conform the DC-OPF problem to a strictly convex QP problem.  See DC-OPF paper
 *
 * To form the input matrices/vector for QuadProgJ, we need following elements:
*-------------------------------
* G:     U, Wrr
* a:     A_1...A_I
* Ceq:   IIk, rBusAdm
* beq:   JJk
* Ciq:   Ot, D, rAdj, Ip, Op
* biq:   bt, bPL, bPU
* ------------------------------
* B:       supplyOffer[][1]
* A:       supplyOffer[][0]
* Wrr:     grid.getReducedVADWeight() ~ Eq(45)
* IIk:     someFunction(atNodeByGen) ~ Eq(54)
* rBusAdm: grid.getReducedBusAdmittance() ~ Eq(51)
* JJk:     someFunction(atNodeByLSE, loadProfile[]) ~ Section 4.3 in DC-OPF paper
* Ot:      NxI zero matrix
* D:       grid.getDiagonalAdmittance() ~ Eq(55)
* rAdj:    grid.getReducedAdjacency() ~ Eq(53)
* Ip:      IxI identity matrix
* Op:      Ix(K-1) zero matrix
* bt:      grid.getLineCap()
* bPL:     supplyOffer[][2]
* bPU:     supplyOffer[][3]

*/
public class DCOPFJ {

  // Index for supplyOffer parameters, i.e., in the form of {a,b,capL,capU}
  private static final int A_INDEX    = 0;
  private static final int B_INDEX    = 1;
  private static final int CAP_LOWER    = 2;
  private static final int CAP_UPPER    = 3;

  private DoubleFactory1D fac1d = DoubleFactory1D.dense;
  private DoubleFactory2D fac2d = DoubleFactory2D.dense;
    // for using Colt's methods e.g. diagonal(), identity(), etc.
  private cern.jet.math.Functions F = cern.jet.math.Functions.functions;
    // F: Naming shortcut to save some keystrokes for calling Colt's functions

  // Input data from BSCUC
  private DoubleMatrix2D supplyOffer;
  private double[] loadProfile;
  private int[] atNodeByGen;
  private int[] atNodeByLSE;
  private TransGrid grid;

  // Intermediate input to form (G,a,Ceq,beq,Ciq,biq)
  private int K;               // numNodes
  private int N;               // numBranches
  private int I;               // numGenAgents
  private int J;               // numLSEAgents
  private DoubleMatrix1D B;    // Ix1
  private DoubleMatrix2D U;    // Ix1
  private DoubleMatrix2D Wrr;  // (K-1)x(K-1)
  private DoubleMatrix1D A;    // Ix1
  private DoubleMatrix2D IIk;  // KxI
  private DoubleMatrix2D rBusAdm;   // (K-1)xK
  private DoubleMatrix1D JJk;  // Kx1
  private DoubleMatrix2D Ot;   // NxI
  private DoubleMatrix2D D;    // NxN
  private DoubleMatrix2D rAdj; // Nx(K-1)
  private DoubleMatrix2D Ip;   // IxI
  private DoubleMatrix2D Op;   // Ix(K-1)
  private DoubleMatrix1D bt;   // Nx1
  private DoubleMatrix1D bPL;  // Ix1
  private DoubleMatrix1D bPU;  // Ix1

  // Input for QuadProgJ
  private DoubleMatrix2D G;
  private DoubleMatrix1D a;
  private DoubleMatrix2D Ceq;
  private DoubleMatrix1D beq;
  private DoubleMatrix2D Ciq;
  private DoubleMatrix1D biq;

  private QuadProgJ qpj;

  // Solution from QuadProgJ
  private double[] commitment;
  private double[] voltAngle;
  private double[] voltAngleDegree;
  private double[] lmp;  // lmp is just eqMultiplier
  private double[] ineqMultiplier;
  private double minTVC; // minTVC = SUM (Ai*PGi + Bi*PGi^2)
  private double[] branchFlow; // branchFlow_km = (1/x_km)(delta_k - delta_m)
  private double[][] bi; // branch index
  private double sumSquaredAngleDifference;  // SUM(delta_k - delta_m)^2

  // Constructor
  public DCOPFJ(double[][] so, double[] lp, int[] ng, int[] nl, TransGrid tg){
    supplyOffer = new DenseDoubleMatrix2D(so);   // Ix4
    loadProfile = lp;  // Jx1
    atNodeByGen = ng;  // Ix1
    atNodeByLSE = nl;  // Jx1
    grid        = tg;

    K = grid.getNumNodes();
    N = grid.getNumBranches();
    I = atNodeByGen.length;
    J = atNodeByLSE.length;

    solveDCOPF();

  }

  private void solveDCOPF(){
    formG();
    forma();
    formCeq();
    formbeq();
    formCiq();
    formbiq();

    qpj = new QuadProgJ(G,a,Ceq,beq,Ciq,biq);

    commitment = new double[I];   // in MWs
    voltAngle = new double[K-1];  // in radians
    voltAngleDegree = new double[K-1]; // in degress
    lmp        = new double[K];
    ineqMultiplier = new double[2*N+2*I];
    minTVC = 0;  // in $/h
    branchFlow = new double[N]; //in MWs
    bi = new double[N][2]; // columns are FROM and TO; rows are branches
    bi = grid.getBranchIndex();
    double[] fullVoltAngle = new double[K]; // including delta_1 = 0
    sumSquaredAngleDifference = 0;


    for(int i=0; i<I; i++){
      commitment[i] = qpj.getMinX()[i]*INIT.getBaseS();
    }
    for(int k=I; k<I+K-1; k++){
      voltAngle[k-I] = qpj.getMinX()[k];  // voltAngle in radians
    }
    for(int k=I; k<I+K-1; k++){
      voltAngleDegree[k-I] = (qpj.getMinX()[k]*180)/Math.PI;  // voltAngle in degrees
    }

    for(int k=0; k<K; k++){
      lmp[k] = qpj.getEqMultipliers()[k]/INIT.getBaseS();
    }
    for(int j=0; j<2*N+2*I; j++){
      ineqMultiplier[j] = qpj.getIneqMultipiers()[j]/INIT.getBaseS();
    }
    for(int i=0; i<I; i++){
      minTVC = minTVC + (A.get(i)/INIT.getBaseS())*commitment[i]
      +(B.get(i)/(INIT.getBaseS()*INIT.getBaseS()))*commitment[i]*commitment[i];
    }

    for(int k=1; k<K; k++){
      fullVoltAngle[k] = voltAngle[k-1];
    }
    
    for(int n=0; n<N; n++){
      branchFlow[n] = (1/grid.getReactance()[n])*(fullVoltAngle[(int)bi[n][0]-1]
                      - fullVoltAngle[(int)bi[n][1]-1])*INIT.getBaseS();

      sumSquaredAngleDifference = sumSquaredAngleDifference
          + Math.pow((fullVoltAngle[(int)bi[n][0]-1]
                      - fullVoltAngle[(int)bi[n][1]-1]),2);
    }

  }

  // G = blockDiag(U,Wr), where U = diag(2B)
  private void formG(){
    B = new DenseDoubleMatrix1D(supplyOffer.viewColumn(B_INDEX).toArray());
    U = new DenseDoubleMatrix2D(fac2d.diagonal(B.assign(F.mult(2))).toArray());
    Wrr = new DenseDoubleMatrix2D(grid.getReducedVADWeight());
    G = new DenseDoubleMatrix2D(I+K-1,I+K-1);
    G.assign(fac2d.composeDiagonal(U,Wrr));
  }

  // a = (A, 0...0)
  private void forma(){
    A = new DenseDoubleMatrix1D(supplyOffer.viewColumn(A_INDEX).toArray());
    a = new DenseDoubleMatrix1D(I+K-1);
    a.viewPart(0,I).assign(A);
  }

  // CeqTranspose = (IIk, -Br'); Ceq = CeqTranspose'; where Br' is rBusAdm here
  private void formCeq(){
    IIk = new DenseDoubleMatrix2D(K,I);
    for(int k=0; k<K; k++){
      for(int i=0; i<I; i++){
        if(atNodeByGen[i]==k+1){
          IIk.set(k,i,1);  // change set to setQuick() later
        }
      }
    }
    rBusAdm = new DenseDoubleMatrix2D(grid.getReducedBusAdmittance());
    DoubleMatrix2D[][] parts = {{ IIk, rBusAdm.viewDice().assign(F.neg)}};
    DoubleMatrix2D CeqTranspose = new DenseDoubleMatrix2D(K,I+K-1);
    CeqTranspose.assign(fac2d.compose(parts));
    Ceq = new DenseDoubleMatrix2D(I+K-1,K);
    Ceq.assign(CeqTranspose.viewDice());
  }
  // JJk = someFunction(atNodeByLSE, loadProfile); beq = JJk
  private void formbeq(){
    JJk = new DenseDoubleMatrix1D(K);
    for(int k=0; k<K; k++){
      double lp=0;
      for(int j=0; j<J; j++){
        if(atNodeByLSE[j]==k+1){
          lp = lp + loadProfile[j];
        }
      }
      JJk.set(k,lp);
    }
    beq = new DenseDoubleMatrix1D(K);
    beq.assign(JJk);
  }
  //CiqTranspose = {{Ot, -Yb*Psir},{Ot, Yb*Psir},{Ip, Op},{-Ip, Op}};
  //Ciq = CiqTranspose'
  // NOTE: Op = -Op
  private void formCiq(){
    Ot   = new DenseDoubleMatrix2D(N,I);
    D    = new DenseDoubleMatrix2D(grid.getDiagonalAdmittance());
    rAdj = new DenseDoubleMatrix2D(grid.getReducedAdjacency());
    Ip = new DenseDoubleMatrix2D(I,I).assign(fac2d.identity(I));
    Op = new DenseDoubleMatrix2D(I,K-1);
    DoubleMatrix2D[][] parts = {
        { Ot,                      D.copy().assign(F.neg).zMult(rAdj,null)},
        { Ot,                      D.zMult(rAdj,null)                     },
        { Ip,                      Op                                      },
        { Ip.copy().assign(F.neg), Op                                      }};
    DoubleMatrix2D CiqTranspose = new DenseDoubleMatrix2D(2*N+2*I,I+K-1);
    CiqTranspose.assign(fac2d.compose(parts));
    Ciq = new DenseDoubleMatrix2D(I+K-1,2*N+2*I);
    Ciq.assign(CiqTranspose.viewDice());
  }
  // biq = (bt, bt, bPL, bPU)
  private void formbiq(){
    bt = new DenseDoubleMatrix1D(grid.getLineCap()).assign(F.neg);
    bPL = new DenseDoubleMatrix1D(supplyOffer.viewColumn(CAP_LOWER).toArray());
    bPU = new DenseDoubleMatrix1D(supplyOffer.viewColumn(CAP_UPPER)
                                  .assign(F.neg).toArray());
    DoubleMatrix1D[] parts = {bt, bt, bPL, bPU};
    biq = new DenseDoubleMatrix1D(2*N+2*I);
    biq.assign(fac1d.make(parts));
  }

  public double[] getCommitment(){
    return commitment;
  }
  public double[] getVoltAngle(){
    return voltAngle;
  }
  public double[] getVoltAngleDegree(){
  return voltAngleDegree;
 }
  public double[] getLMP(){
    return lmp;
  }
  public double[] getIneqMultiplier(){
    return ineqMultiplier;
  }
  public double getMinTVC(){
    return minTVC;
  }
  public double[] getBranchFlow(){
    return branchFlow;
  }
  public double getSumSquaredAngleDifference(){
    return sumSquaredAngleDifference;
  }
  public int getNumBindingConstraints(){
    return qpj.getNumBC();
  }
  public int[] getActiveSet(){
    return qpj.getActiveSet();
  }
  public boolean getIsSolutionFeasibleAndOptimal(){
    return qpj.getIsFeasibleAndOptimal();
  }

  public static void main(String[] args) {
    
    System.out.println("<><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><>\n");
    System.out.println("     DCOPFJ: An Open-Source Java Solver for DC Optimal Power Flow (OPF) Problems    \n");
    System.out.println("             (C) Copyright 2007, by Junjie Sun and Leigh Tesfatsion              ");
    System.out.println("                            Version Release: " + 1.1);    
    System.out.println("        Package Site: http://www.econ.iastate.edu/tesfatsi/DCOPFJHome.htm");
    System.out.println("         Licensed as free open-source software by the copyright holders ");
    System.out.println("    (Sun and Tesfatsion) under the terms of the GNU General Public License (GPL) \n");
    System.out.println("<><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><>\n");    
    
    String[] caseName = {"3NodeCase","5NodeCase"};
    // casdID: 0 = 3NodeCase, 1 = 5NodeCase
    int caseID = 1;

    INIT init = new INIT("DATA/" + caseName[caseID] +".txt");
    System.out.println("\nCase name: " + caseName[caseID] + "\n");


    TransGrid grid = new TransGrid(init.getNodeData(),init.getBranchData(),40,40);
    double[][] genData = init.getGenData();
    double[][] lseData = init.getLSEData();
    int K = grid.getNumNodes();
    int N = grid.getNumBranches();
    int I = genData.length;
    int J = lseData.length;
    int H = 24;

    double[][] loadProfile = new double[J][H];

    DCOPFJ opf;
    double[][] dailyCommitment = new double[H][I]; // (24 hours by row)
    double[][] dailyVoltAngle = new double[H][K];
    double[][] dailyLMP = new double[H][K];
    double[][] dailyIneqMultiplier = new double[H][2*N+2*I];
    double[] dailyMinTVC = new double [H];
    double[][] dailyBranchFlow = new double[H][N];
    double[] dailySumSquaredAngleDifference = new double[H];

    for(int j=0; j<J; j++){
      for(int h=0; h<H; h++){
        loadProfile[j][h] = lseData[j][h+2];
      }
    }

    int[] atNodeByGen = new int[I];
    for(int i=0; i<I; i++){
      atNodeByGen[i] = (int) genData[i][1];
    }

    int[] atNodeByLSE = new int[J];
    for(int j=0; j<J; j++){
      atNodeByLSE[j] = (int) lseData[j][1];
    }

    double[][] supplyOfferByGen = new double[I][4];
    for(int i=0; i<I; i++){
      for(int p=0; p<4; p++){
        supplyOfferByGen[i][p] = genData[i][p+3];
      }
    }

    for (int h=0; h<H; h++){
      

      double[] hourlyLoadProfileByLSE = new double[J];
      for(int j=0; j<J; j++){
        hourlyLoadProfileByLSE[j] = loadProfile[j][h];
      }

      opf = new DCOPFJ(supplyOfferByGen,hourlyLoadProfileByLSE,
                atNodeByGen, atNodeByLSE, grid);

      dailyCommitment[h] = opf.getCommitment();
      dailyVoltAngle[h] = opf.getVoltAngleDegree();//report voltAngle in degrees
      dailyLMP[h]        = opf.getLMP();
      dailyIneqMultiplier[h] = opf.getIneqMultiplier();
      dailyMinTVC[h] = opf.getMinTVC();
      dailyBranchFlow[h] = opf.getBranchFlow();
      dailySumSquaredAngleDifference[h] = opf.getSumSquaredAngleDifference();
    
    }

    // Extracting capL and capU from the 3rd and 4th column of supplyOfferByGen 
    //  (used to show if the lower or upper constraints are binding in solution)
    double[] capL = new double[I];
    double[] capU = new double[I];
    for(int i=0; i<I; i++){
      capL[i] = supplyOfferByGen[i][2]*INIT.getBaseS();
    }
    for(int i=0; i<I; i++){
      capU[i] = supplyOfferByGen[i][3]*INIT.getBaseS();
    }
    
    System.out.println("  Daily 24-hour generation commitments (in MWs) for each Generator");
    Support.print("Gen",I);
    Support.printWithContraints(dailyCommitment,2,capL,capU);
    
    
    System.out.println("  Daily 24-hour voltage angles (in radians) for each non-reference node");
    Support.printNonRef("Delta",K);
    Support.print(dailyVoltAngle,4);
  
    System.out.println("  Daily 24-hour locational marginal prices (in $/MWh) for each node");
    Support.print("LMP", K);
    Support.print(dailyLMP,2);
    System.out.println("  Daily 24-hour minimum total variable cost levels (in $/h) for the system");
    Support.print(dailyMinTVC,2);
    System.out.println("  Daily 24-hour branch flows (in MWs) for each branch, in the order of bi");
    Support.print("BFlow", N);
    Support.print(dailyBranchFlow,2);
    System.out.println("  Daily 24-hour inequality contraint multipliers, in the order of biq");
    Support.printArray2dp(dailyIneqMultiplier);
    System.out.println("  Daily 24-hour sums of squared voltage angle differences (in radians)");
    Support.printArray(dailySumSquaredAngleDifference);    




  }

}
