/* ============================================================================
 * DCOPFJ : An open-source Java solver for DC optimal power flow (OPF) problems
 * ============================================================================
 * 
 * (C) Copyright 2007, by Junjie Sun and Leigh Tesfatsion
 *
 * Version 1.1
 *
 * Project site:  http://www.econ.iastate.edu/tesfatsi/DCOPFJHome.htm
 *
 * LICENSING TERMS
 * The DCOPFJ package is licensed by the copyright holders (Junjie Sun and 
 * Leigh Tesfatsion) as free open-source software under the terms of the GNU
 * General Public License (GPL). Anyone who is interested is allowed to view, 
 * modify, and/or improve upon the code used to produce this package, but any 
 * software generated using all or part of this code must be released as free 
 * open-source software in turn. The GNU GPL can be viewed in its entirety as 
 * in the following site: http://www.gnu.org/licenses/gpl.html
 */

package dcopfj;

// INIT.java
// Initialization of input data

import java.io.*;
import java.util.StringTokenizer;
import java.util.ArrayList;
import cern.colt.matrix.*;
import cern.colt.matrix.impl.*;

/**
* Example of initial data file format in SI
*
NAME	3NodeCase
UNIT	SI
BASE_S	100
BASE_V	10


// SI (International System of Units) such as MW
// PU (Per Unit)

// Source: Shahidehpour, Yamin and Li (2002), "Market Operations in Electric
//         Power Systems: Forecasting, Scheduling, and Risk Management,"
//         John Wiley, IEEE, page 297, Table 8.2-8.4

NodeData
//NN	penaltyCoeff
  3	0.05

BranchData
//From	To	lineCap	reactance
  1	2	55	0.20
  1	3	55	0.40
  2	3	55	0.25

GenData
//ID	atNode	FCost	a	b	capL	capU	initMoney
  1	1	142.735	10.694	0.00463	20	200	10000
  2	2	218.335	18.1	0.00612	10	150	10000
  3	3	118.821	37.8896	0.01433	5	20	10000

LSEData
//ID	atNode	LP-00	LP-01	LP-02	LP-03	LP-04	LP-05	LP-06	LP-07 
  1	1	132.66	122.4	115.62	112.2	108.84	110.52	112.2	119.04
  2	2	44.22	40.8	38.54	37.4	36.28	36.84	37.4	39.68
  3	3	44.22	40.8	38.54	37.4	36.28	36.84	37.4	39.68

//ID	atNode	LP-08	LP-09	LP-10	LP-11	LP-12	LP-13	LP-14	LP-15
  1	1	136.02	149.64	153.06	154.74	153.06	149.64	147.96	147.96
  2	2	45.34	49.88	51.02	51.58	51.02	49.88	49.32	49.32
  3	3	45.34	49.88	51.02	51.58	51.02	49.88	49.32	49.32

//ID	atNode	LP-16	LP-17	LP-18	LP-19	LP-20	LP-21	LP-22	LP-23 
  1	1	154.74	170.04	163.26	161.52	159.84	156.42	147.96	137.76
  2	2	51.58	56.68	54.42	53.84	53.28	52.14	49.32	45.92
  3	3  	51.58	56.68	54.42	53.84	53.28	52.14	49.32	45.92

EndOfFile

// NN: Number of Nodes
// penaltyCoeff: soft penalty coefficient pi
// lineCap: Thermal limits
// capL,capU: Generator's operating capacity limits
// FCost,a,b: Generator's cost attributes
// LP-H: load profile (fixed demand) at hour H, where x=00,01,...,23

NOTE: any line starts with // will be ignored when data are read in

 */

public class INIT {

  // Tolerance value (used in correctRoundingError())
  private static final double TOL = 1.0E-9;

  // Default values
  public static final int BASE_S = 100;
  public static final int BASE_V = 10;

  // Number of parameters (number of columns in the data file)
  public static final int NUM_NODE_PARAMS = 2;
    // numNodes, softPenaltyCoeff
  public static final int NUM_BRANCH_PARAMS = 4;
    // FromNode, ToNode, reactance, lineCap (Thermal Limits)
  public static final int NUM_GEN_PARAMS = 8;
    // genID, atWhichNode, FCost, a, b, capL, capU, initMoney
  public static final int NUM_LSE_3SEC_PARAMS = 2+8;
    // lseID, atWhichNode, LP-01,...,LP-08
  public static final int NUM_LSE_PARAMS = 2+24;
    // lseID, atWhichNode, LP-01,...,LP-24

  private String[][] paramNames = {
   {"NN","penaltyCoeff"},
   {"From","To","lineCap","reactance"},
   {"ID","atNode","FCost","a","b","capL","capU","initMoney"},
   {"ID","atNode","LP-00","LP-01","LP-02","LP-03","LP-04","LP-05","LP-06","LP-07",
                  "LP-08","LP-09","LP-10","LP-11","LP-12","LP-13","LP-14","LP-15",
                  "LP-16","LP-17","LP-18","LP-19","LP-20","LP-21","LP-22","LP-23"}};

  private double[][] nodeData;
  private double[][] branchData;
  private double[][] genData;
  private double[][] lse3SecData;  // 3-sectional 8-hour LSE data
  private double[][] lseData;      // Combine 3-sectional to 24-hour LSE data
  private boolean isPU;
  private static double baseS;
  private static double baseV;

  private ArrayList slist = new ArrayList();
    //slist: the master string list to store tokenized strings in data file

  // Constructor
  public INIT(String initFilePath) {

    //Do an initial default load in case some parameters are missing in the configuration file.
    loadDefaultConfig();
    try {
      readInitData(initFilePath);
    }
    catch (Exception ioe) {loadDefaultConfig();
      System.err.println(
          "Problem loading simulation configurtation from file. " +
          "Using default configuration.");
    }
  }

  private void readInitData(String initFilePath){

    try{

      // Create a stream tokenizer by creating a buffer reader and a file reader
      FileReader       fr = new FileReader(initFilePath);
      BufferedReader   br = new BufferedReader(fr);
      StreamTokenizer tok = new StreamTokenizer(br);

      // Resets tokenizer's syntax table so that all characters are "ordinary"
      // to override Java's default setting for tokenization, which is too
      // finely tuned.
      tok.resetSyntax();

      // Set characters in the ASCII range 33 to 255 as valid word characters
      tok.wordChars(33,255);

      // Set characters in the ASCII range 0 to 32(' ') as whitespace characters
      tok.whitespaceChars(0,' ');   // or: tok.whitespaceChars(0,32);

    //Specified that the character argument starts a single-line comment.
      tok.commentChar('/');

      // Add tokens to a String list, slist
      while(tok.nextToken() != tok.TT_EOF){
        slist.add(tok.sval);
      }

      // Parse slist to double[][] arrays, The 'main' method
      parseStringList();

      //Close file reader
      fr.close();

    }catch(Exception e){
    System.out.println("Exception: " + e);
  }
  }

  private void parseStringList(){
    setNodeData();
    setBranchData();
    setGenData();
    setLSEData();
    int index = slist.indexOf("UNIT")+1;
    isPU = slist.get(index).toString().equalsIgnoreCase("PU");
    if(isPU == false){
      convertToPU();
    }

  }

  private double[][] correctRoundingError(double[][] a){
    for(int i=0; i<a.length; i++){
      for(int j=0; j<a[0].length; j++){
        for(int p=1; p<1.0E6; p=p*10){
          if(Math.abs(a[i][j]*p-Math.rint(a[i][j]*p)) < TOL){
            a[i][j] = Math.rint(a[i][j]*p)/p;
            break;
          }
        }
      }
    }
    return a;
  }

  private double correctRoundingError(double scalar){
    for(int p=1; p<1.0E6; p=p*10){
      if(Math.abs(scalar*p-Math.rint(scalar*p)) < TOL){
        scalar = Math.rint(scalar*p)/p;
        break;
      }
    }
    return scalar;
  }


  private void convertToPU(){

    boolean isThereBaseS = slist.contains((String) "BASE_S");
    boolean isThereBaseV = slist.contains((String) "BASE_V");

    if(isThereBaseS == true){
      int index = slist.indexOf("BASE_S")+1;
      baseS = Double.parseDouble(slist.get(index).toString());
    }
    else{
      baseS = BASE_S;
    }
    if(isThereBaseV == true){
      int index = slist.indexOf("BASE_V")+1;
      baseV = Double.parseDouble(slist.get(index).toString());
    }
    else{
      baseV = BASE_V;
    }

    for(int n=0; n<branchData.length; n++){
      // Convert lineCap from SI to PU
      branchData[n][2] = branchData[n][2]/baseS;
      // Convert reactance from SI to PU, x(pu) = x/Zo = x/(Vo^2/So) = (x*So)/Vo^2
      branchData[n][3] = (branchData[n][3]*baseS)/(baseV*baseV);
    }
    for(int i=0; i<genData.length; i++){
      genData[i][3] = genData[i][3]*baseS;       // Convert A from SI to PU-adjusted
      genData[i][4] = genData[i][4]*baseS*baseS; // Convert B from SI to PU-adjusted
      genData[i][5] = genData[i][5]/baseS;       // Convert capL from SI to PU
      genData[i][6] = genData[i][6]/baseS;       // Convert capU from SI to PU
    }
    for(int j=0; j<lseData.length; j++){
      for(int h=2; h<lseData[0].length; h++){
        lseData[j][h] = lseData[j][h]/baseS;     // Convert LP from SI to PU
      }
    }
    branchData = correctRoundingError(branchData);
    genData    = correctRoundingError(genData);
    lseData    = correctRoundingError(lseData);
  }

  private void setNodeData(){
    int fromIndex = slist.indexOf("NodeData")+1;
    int   toIndex = slist.indexOf("BranchData");
    int numValues = slist.subList(fromIndex,toIndex).size();
       //NOTE: fromIndex is inclusive, and toIndex is exclusive for subList().
    nodeData = new double[numValues/NUM_NODE_PARAMS][NUM_NODE_PARAMS];
    parseToArray2D(nodeData, NUM_NODE_PARAMS, fromIndex, toIndex);
  }

  private void setBranchData(){
    int fromIndex = slist.indexOf("BranchData")+1;
    int   toIndex = slist.indexOf("GenData");
    int numValues = slist.subList(fromIndex,toIndex).size();
    branchData = new double[numValues/NUM_BRANCH_PARAMS][NUM_BRANCH_PARAMS];
    parseToArray2D(branchData, NUM_BRANCH_PARAMS, fromIndex, toIndex);
  }

  private void setGenData(){
    int fromIndex = slist.indexOf("GenData")+1;
    int   toIndex = slist.indexOf("LSEData");
    int numValues = slist.subList(fromIndex,toIndex).size();
    genData = new double[numValues/NUM_GEN_PARAMS][NUM_GEN_PARAMS];
    parseToArray2D(genData, NUM_GEN_PARAMS, fromIndex, toIndex);
  }

  private void setLSEData(){
    int fromIndex = slist.indexOf("LSEData")+1;
    int   toIndex = slist.indexOf("EndOfFile");
    int numValues = slist.subList(fromIndex,toIndex).size();

    // Create lse3SecData as double[90/10][10] or 9-by-10 2D array
    int row3 = numValues/NUM_LSE_3SEC_PARAMS;    // row3 = 90/10 = 9
    int col3 = NUM_LSE_3SEC_PARAMS;              // col3 = 10
    lse3SecData = new double[row3][col3];
    parseToArray2D(lse3SecData, NUM_LSE_3SEC_PARAMS, fromIndex, toIndex);

    // Create lseData as double[90/(3*10)][26] or 3-by-26 2D array
    int row = numValues/(3*NUM_LSE_3SEC_PARAMS); // row = 90/(3*10) = 3
    int col = NUM_LSE_PARAMS;                    // col = 26
    lseData = new double[row][col];

    DoubleMatrix2D lse3SecDataDM = new DenseDoubleMatrix2D(lse3SecData);
    DoubleMatrix2D lseDataDM = new DenseDoubleMatrix2D(lseData);
    lseDataDM.viewPart(0,0,row,col3)
        .assign(lse3SecDataDM.viewPart(0,0,row,col3));
    lseDataDM.viewPart(0,col3,row,col3-2)
        .assign(lse3SecDataDM.viewPart(row,2,row,col3-2));
    lseDataDM.viewPart(0,2*col3-2,row,col3-2)
        .assign(lse3SecDataDM.viewPart(2*row,2,row,col3-2));
    lseData = lseDataDM.copy().toArray();
  }

  private void parseToArray2D(double[][] data, int numParams,
                            int fromIndex, int toIndex){
    int row = 0;
    int col = -1;
    for(int i=fromIndex; i<toIndex; i++){
      if(col >= numParams-1){
        col = 0;
        row++;
      }
      else{
        col++;
      }
      data[row][col] = Double.parseDouble(slist.get(i).toString());
    }
  }

  private void loadDefaultConfig(){

  }

  public String[][] getParamNames(){
    return paramNames;
  }
  public static double getBaseS(){
    return baseS;
  }
  public static double getBaseV(){
     return baseV;
   }

  public double[][] getNodeData(){
    return nodeData;
  }

  public double[][] getBranchData(){
    return branchData;
  }
  public double[][] getGenData(){
    return genData;
  }
  public double[][] getLSEData(){
    return lseData;
  }


  public static void printArray(double [][]a){
    for(int i = 0; i<a.length; i++){
      for(int j = 0 ; j<a[0].length; j++){
        System.out.print("\t" + a[i][j]);
        }
        System.out.println();
      }
      System.out.println();
  }

  public static void printArray(String[] s){
    for(int i=0; i<s.length; i++){
      System.out.print("\t" + s[i]);
    }
    System.out.println();
  }  

}
